/*
 *    snake - a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * file: ui_X.c *  purpose: handle X graphics 
 */

#include "sins.h"
#include "ui_picts.h"
#include <X11/Xlib.h> 

#define USE_ROOT_WINDOW

/* public functions */
int ui_init ();
int ui_initarena (ARENA *);
void ui_finish (void);
void ui_drawarena (ARENA *);
void ui_drawmessage (char *);
char *ui_prompt (char *);
int ui_getkey (void);
int CMD_loadpictset (char *line);

/* private functions */
static void init_background(ARENA *);
static void draw_pict (Drawable w, int sx, int sy, int pictnum);
static int *load_pictset (char *filename);
#ifndef USE_ROOT_WINDOW
static Window create_arenawin (Display *dpy, int width, int height);
#endif
static void create_pictsetwin (void);
static void enable_backingstore(Display *d, Window w);
static int send_pictset (void);
static int copy_pict (int x, int y, int pictnum);

/* private data */
static ARENA oldarena;
static int spriteheight;
static int spritewidth;
static int *pictset;
static int pictsetsent = 0;

static Display *dpy;
static Screen *scr;
static Window arenawin;
static Window pictsetwin;
static GC gc; 


/* public data */
int screen_initialized = 0;

/* Initialize UI module */
int
ui_init ()
{
  if (!register_command ("loadpictset", &CMD_loadpictset,
			 "loadpictset [<fname>]")) return 0;

  pictset = defpictset;

  return 1;
}

/* initialize the graphics */
int
ui_initarena (ARENA *arena)
{
  int ydim, xdim;

  if ( ! (dpy=XOpenDisplay(NULL)) ) 
  {
    message ("Can't open display: %s", getenv("DISPLAY"));
    finish();
  }

  scr = DefaultScreenOfDisplay(dpy);

  /* default dimensions */
#ifdef USE_ROOT_WINDOW
  xdim = scr->width;
  ydim = scr->height;
#else
  xdim = (scr->width)/2;
  ydim = (scr->height)/2;
#endif

  /* Set window dimensions based on arena and sprites dimensions */
  if ( !spritewidth ) spritewidth=8;
  if ( !spriteheight ) spriteheight=8;
  if ( arena->lines )
  {
    /* set window height based on arena and sprite heights */
    ydim = spriteheight * ( arena->lines );
  }
  else
  {
    /* set arena height based on window and sprite heights */
    arena->lines = ydim/spriteheight;
  }
  if ( arena->cols )
  {
    /* set window width based on arena and sprite widths */
    xdim = spritewidth * ( arena->cols );
  }
  else
  {
    /* set arena width based on window and sprite widths */
    arena->cols = xdim/spritewidth;
  }

#ifdef USE_ROOT_WINDOW
  arenawin = DefaultRootWindow(dpy);
#else
  arenawin = create_arenawin(dpy, xdim, ydim);
#endif
  enable_backingstore(dpy, arenawin);

  gc = XCreateGC(dpy, arenawin, 0, NULL);

  create_pictsetwin();

  send_pictset();

  init_background(arena);

  XFlush(dpy);

  /* initialize the oldarena (used for optimization purposes) */
  oldarena.lines = arena->lines;
  oldarena.cols = arena->cols;
  oldarena.init = 1; /* useless ? */
  oldarena.board = (sprite **)safe_malloc(
    oldarena.lines*oldarena.cols*sizeof(sprite),
    "ui_X: ui_drawarena: oldarena.board");
  memset(oldarena.board, 0, oldarena.lines*oldarena.cols);

  /* Set screen initialized flag */
  screen_initialized = 1;

  return 1;

}

/* X version of message drawing function */
void
ui_drawmessage (char *str)
{

  if ( strlen(str) ) fprintf(stderr, "%s\n", str);

#ifndef USE_ROOT_WINDOW
  /* This is pretty stupid isn't it ? */
  if ( strlen(str) ) XStoreName (dpy, arenawin, str);
  else XStoreName (dpy, arenawin, "X-Sins");
#endif

  return;
}

void
ui_drawarena (ARENA *arena)
{
  int y, x;
  sprite *sp;
  int pictidx;
  int should_redraw = 0;

  /* Send pictset to the display if needed */
  if ( ! pictsetsent )
  {
    send_pictset();
    should_redraw = 1;
  }

  for (y = 0; y < arena->lines; y++)
  {
    for (x = 0; x < arena->cols; x++)
    {

      sp = GETSPRITE(arena,y,x);

      /* no changes since last drawing */
      if ( ! should_redraw ) if ( (sp==GETSPRITE(&oldarena,y,x)) ) continue;

      /* update oldarena struct */
      oldarena.board[y*oldarena.cols+x] = sp;

      if (sp) switch (sp->type)
      {
        case SNAKE:
          pictidx = PICT_SNAKE0+((snakebody *)(sp))->self->playernum;
          break;  
        case EXSNAKE:
          pictidx = PICT_EXSNAKE;
          break; 
        case FRUIT:
          pictidx = PICT_FRUIT0+(((fruit *)(sp))->value)-48;
          break;
        case WALL:
          pictidx = PICT_WALL;
          break;
        default:
          pictidx = PICT_UNKNOWN;
          break;
      }
      else pictidx = PICT_EMPTY;

      /* draw_pict(arenawin, x, y, pictidx); */
      copy_pict(x, y, pictidx);
    }
  }

  XFlush(dpy);
}


void
ui_finish ()
{
  if ( screen_initialized )
  {
    XCloseDisplay(dpy);
  }
}


/*
 * Get next key in input buffer
 */
int
ui_getkey ()
{
  int key;

#ifndef USE_ROOT_WINDOW_test
  XEvent event;

  if ( XCheckWindowEvent(dpy, arenawin, KeyPressMask, &event) )
  {
    key = XKeycodeToKeysym(dpy, event.xkey.keycode, 0);

    /* I found this to be working :) */
    if ( key >= 128 ) key = key%128;

    /* .. just don't know how to handle Capital letters */

    return key;
  }
#endif

  return -1;
}

/*
 * prompt the user for a string
 * the line returnered is allocated with malloc(3)
 * so the caller must free it when finished with it. 
 */
char *
ui_prompt (char *prompt_string)
{
  return NULL;
}

int
CMD_loadpictset (char *line)
{
  char fname[256];
  int ret = 0;
  int *newset;

  if ( line ) ret = sscanf (line, "%s", fname);
  if ( ret != 1 )
  {
    /* set default pictset (should instead complain about wrong num of args?) */
    pictset = defpictset;
    message("default pictset loaded");
    pictsetsent = 0;
    return 1;
  }

  newset = load_pictset(fname);
  if ( !newset )
  {
    message("Can't load pictset from %s", fname);
    return 0;
  }

  if ( pictset != defpictset ) free(pictset);
  pictset = newset;

  pictsetsent = 0;

  message("pictset %s loaded", fname);
  return 1;
}


/************ PRIVATE FUNCTIONS **********************/


static void
enable_backingstore (Display *d, Window w)
{
  XSetWindowAttributes wa;
  wa.backing_store = Always;
  XChangeWindowAttributes(d, w, CWBackingStore, &wa);
}

#ifndef USE_ROOT_WINDOW
static Window
create_arenawin (Display *dpy, int width, int height)
{
  Window w;

  w = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
    width, height, 0, 0, 0);

  /* We want to get MapNotify, KeyPress and Expose events */
  XSelectInput(dpy, w, StructureNotifyMask|KeyPressMask);

  /* "Map" the window (that is, make it appear on the screen) */
  XMapWindow(dpy, w);

  for(;;)
  {
    XEvent e;
    XNextEvent(dpy, &e);
    if (e.type == MapNotify) break;
  }

  /* Give window a name */
  XStoreName (dpy, w, "X-Sins");

  return w;
}
#endif


static void 
create_pictsetwin (void)
{

  pictsetwin = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
    spritewidth, spriteheight*PICTSETLEN, 0, 0, 0);

  enable_backingstore(dpy, pictsetwin);

}

static void
init_background(ARENA *arena)
{
  int y,x;

  /* initialize background */
  for (y = 0; y < arena->lines; y++)
  {
    for (x = 0; x < arena->cols; x++)
    {
        /*draw_pict(arenawin, x, y, PICT_EMPTY);*/
        copy_pict(x, y, PICT_EMPTY);
    }
  }
}

static void
draw_pict (Drawable w, int sx, int sy, int pictnum)
{
  int x,y,i,j;
  int *pict;
  int color;

  sx=sx*spritewidth;
  sy=sy*spriteheight;
  pict=pictset+pictnum*spritewidth*spriteheight;

  for (x=sx, i=0; i<spritewidth; i++, x++)
  {
    for (y=sy, j=0; j<spriteheight; y++, j++)
    {
      color = *(pict+(j*spritewidth)+i);
      XSetForeground( dpy, gc, color );
      XDrawPoint( dpy, w, gc, x, y );
    }
  }
}

static int
copy_pict (int x, int y, int pictnum)
{
  int sx, sy, dx, dy;

  sx = 0;
  sy = pictnum*spriteheight;
  dx=x*spritewidth;
  dy=y*spriteheight;

  XCopyArea(dpy,pictsetwin,arenawin,gc,sx,sy,spritewidth,spriteheight,dx,dy);

  return 1;
}



static int *
load_pictset (char *filename)
{
  FILE *stream;
  int idx = 0;
  int quoted = 0;
  int ch;
  int *set;
  int w, h;

  stream = fopen(filename, "r");
  if ( !stream ) return NULL;

  /* check sprite size */
  if ( fscanf(stream, "#%ix%i\n", &w, &h) != 2 )
  {
    message("Invalid pictset format (missing '#WxH' on line 1)");
    return NULL;
  }

  if ( screen_initialized && (w != spritewidth || h != spriteheight) )
  {
    message("Can't change sprite size while running");
    return NULL;
  }

  spritewidth = w;
  spriteheight = h;
  set = safe_malloc(w*h*PICTSETLEN*sizeof(int), "New pictset");

  while (!feof(stream))
  {
    if ( idx >= w*h*PICTSETLEN ) break;
    ch=getc(stream);
    if (ch) switch(ch)
    {
      case '\n':
      case '\r':
        quoted=0;
        break;
      case ' ':
      case '\t':
        break;
      case '#':
        quoted=1;
        break;
      default:
        if ( !quoted )
        {
          *(set+idx) = ch-97;
          idx++;
        }
        break;
    }
  }
  fclose(stream);
  return set;
}

static int
send_pictset()
{
  int n;

  for (n=0; n<PICTSETLEN; n++)
  {
    draw_pict(pictsetwin, 0, n, n);
  }

  message("Pictset sent");
  pictsetsent = 1;

  return 1;
}

